from typing import TYPE_CHECKING, Any, Dict, Optional, Type, TypeVar, Union

import pydantic
from pydantic import schema_of
from pydantic.error_wrappers import ErrorWrapper, ValidationError
from pydantic.main import ROOT_KEY
from pydantic.tools import NameFactory, parse_obj_as
from simdjson import Parser

from simdjson_schemaful import loads
from simdjson_schemaful.parser import Schema

if TYPE_CHECKING:
    Model = TypeVar("Model", bound="BaseModel")


class ModelMetaclass(pydantic.main.ModelMetaclass):
    def __new__(cls, *args: Any, **kwargs: Any) -> type:
        ret = super().__new__(cls, *args, **kwargs)
        _REGISTRY[ret] = ret.schema()
        return ret


T = TypeVar("T")
_REGISTRY: Dict[ModelMetaclass, Schema] = {}


class BaseModel(pydantic.BaseModel, metaclass=ModelMetaclass):
    @classmethod
    def parse_raw_simdjson(
        cls: Type["Model"],
        b: Union[str, bytes],
        parser: Optional[Parser] = None,
    ) -> "Model":
        try:
            obj = loads(b, schema=_REGISTRY[cls], parser=parser)
        except (ValueError, TypeError, UnicodeDecodeError) as e:
            raise ValidationError([ErrorWrapper(e, loc=ROOT_KEY)], cls)
        return cls.parse_obj(obj)


def parse_raw_simdjson_as(
    type_: Type[T],
    b: Union[str, bytes],
    *,
    parser: Optional[Parser] = None,
    type_name: Optional[NameFactory] = None,
    **_: Any,
) -> T:
    schema = schema_of(type_)  # already cached in pydantic
    obj = loads(b, schema=schema, parser=parser)
    return parse_obj_as(type_, obj, type_name=type_name)



import json
from typing import TYPE_CHECKING, Any, Dict, Generic, Optional, Type, TypeVar, Union

import pydantic
from pydantic import ValidationError
from pydantic_core import InitErrorDetails, PydanticCustomError
from simdjson import Parser

from simdjson_schemaful import loads
from simdjson_schemaful.parser import Schema

if TYPE_CHECKING:
    Model = TypeVar("Model", bound="BaseModel")


class ModelMetaclass(pydantic._internal._model_construction.ModelMetaclass):
    def __new__(cls, *args: Any, **kwargs: Any) -> type:
        ret = super().__new__(cls, *args, **kwargs)
        _REGISTRY[ret] = ret.model_json_schema()  # type: ignore
        return ret


T = TypeVar("T")
_REGISTRY: Dict[ModelMetaclass, Schema] = {}


class BaseModel(pydantic.BaseModel, metaclass=ModelMetaclass):
    @classmethod
    def model_validate_simdjson(
        cls: Type["Model"],
        json_data: Union[str, bytes, bytearray],
        parser: Optional[Parser] = None,
    ) -> "Model":
        try:
            obj = loads(json_data, schema=_REGISTRY[cls], parser=parser)
        except (ValueError, TypeError, UnicodeDecodeError) as e:
            raise ValidationError(e)
        return cls.model_validate(obj)


class TypeAdapter(Generic[T]):
    __slots__ = ("_ta", "_simdjson_schema")

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._ta = pydantic.TypeAdapter[T](*args, **kwargs)
        self._simdjson_schema = self._ta.json_schema()

    @property
    def pydantic_type_adapter(self) -> pydantic.TypeAdapter[T]:
        return self._ta

    def _build_error(self, exc: Exception, data: Union[str, bytes]) -> ValidationError:
        if isinstance(exc, UnicodeDecodeError):
            type_str = "value_error.unicodedecode"
        elif isinstance(exc, json.JSONDecodeError):
            type_str = "value_error.jsondecode"
        elif isinstance(exc, ValueError):
            type_str = "value_error"
        else:
            type_str = "type_error"

        details: InitErrorDetails = {
            "type": PydanticCustomError(type_str, str(exc)),
            "loc": ("__root__",),
            "input": data,
        }
        return ValidationError.from_exception_data(
            self._ta.core_schema["type"],
            [details],
        )

    def validate_simdjson(
        self,
        data: Union[str, bytes],
        *,
        strict: Optional[bool] = None,
        context: Optional[Dict[str, Any]] = None,
        parser: Optional[Parser] = None,
    ) -> T:
        try:
            obj = loads(data, schema=self._simdjson_schema, parser=parser)
        except (ValueError, TypeError, UnicodeDecodeError) as e:
            raise self._build_error(e, data)
        return self._ta.validate_python(obj, strict=strict, context=context)



from importlib.util import find_spec

from pydantic import BaseModel

if find_spec("pydantic.v1"):
    from .v2 import TypeAdapter

    __all__ = (
        "BaseModel",
        "TypeAdapter",
    )
else:
    from .v1 import parse_raw_simdjson_as

    __all__ = (
        "BaseModel",
        "parse_raw_simdjson_as",
    )



from typing import Any, Callable, Dict, List, Optional, Union

import simdjson
from simdjson import Parser

JsonType = Union[Dict[Any, Any], List[Any], str, int, float, bool]
Schema = Dict[Any, Any]
_Dict = Dict[Any, Any]
_List = List[Any]
_FuncSet = Callable[..., None]


# TODO: handle anyOf?


def _get_definition(definitions: Dict[str, Schema], schema: Schema) -> Schema:
    if ref := schema.get("$ref"):
        return definitions[ref.split("/")[-1]]
    return schema


def _set_dict(target: _Dict, key: str, value: Any) -> None:
    target[key] = value


def _set_list(target: _List, index: int, value: Any) -> None:
    if len(target) <= index:
        target.extend([None] * (index + 1 - len(target)))
    target[index] = value


def _process_prop(
    *,
    prop_data: _Dict,
    prop: Union[str, int],
    value: Any,
    target: Union[_Dict, _List],
    func_set: _FuncSet,
    definitions: Schema,
    queue: _List,
) -> None:
    if value is None:
        target[prop] = value  # type: ignore
        return

    type_ = prop_data.get("type")
    items = prop_data.get("items", {})

    if type_ == "array" and not items.get("$ref") and not items.get("properties"):
        if not isinstance(value, simdjson.Array):
            raise ValueError(
                f"Supposed to be an array, but in reality is a {value.__class__}",
            )
        func_set(target, prop, value.as_list())
        return

    if (not type_ and not prop_data.get("$ref")) or (
        type_ == "object"
        and not prop_data.get("properties")
        and not prop_data.get("additionalProperties")
    ):
        if not isinstance(value, simdjson.Object):
            raise ValueError(
                f"Supposed to be an object, but in reality is a {value.__class__}",
            )
        func_set(target, prop, value.as_dict())
        return

    if type_ not in (None, "array", "object"):
        if isinstance(value, (simdjson.Object, simdjson.Array)):
            raise ValueError(
                f"Supposed to be anything but object/array, "
                f"but in reality is {value.__class__}",
            )
        func_set(target, prop, value)
        return

    if type_ == "array":
        definition = prop_data
        func_set(target, prop, [])
    elif type_ == "object" or prop_data.get("$ref"):
        definition = _get_definition(definitions, prop_data)
        func_set(target, prop, {})
    else:
        raise ValueError(f"invalid type {type_} for prop data {prop_data}")

    queue.append((definition, value, target[prop]))  # type: ignore


def _loads(  # noqa: C901
    data: Union[bytes, bytearray, memoryview],
    *,
    schema: Schema,
    parser: Parser,
) -> JsonType:
    definitions = schema.get("definitions", {}) or schema.get("$defs", {})

    if "$ref" in schema:
        schema = _get_definition(definitions, schema)

    type_ = schema.get("type")
    if type_ not in ["object", "array"]:
        return simdjson.loads(data)  # type: ignore

    res: Union[_List, _Dict] = {} if type_ == "object" else []

    source = parser.parse(data)
    target = res
    queue = [(schema, source, target)]

    while queue:
        schema, source, target = queue.pop()
        type_ = schema["type"]

        if type_ == "object":
            if not isinstance(source, simdjson.Object):
                raise ValueError(
                    f"Supposed to be an object, but in reality is a {source.__class__}",
                )

            properties = schema.get("properties", {})

            if properties:
                for prop_name, prop_data in properties.items():
                    value = None
                    try:
                        value = source[prop_name]
                    except KeyError:
                        continue

                    _process_prop(
                        prop_data=prop_data,
                        prop=prop_name,
                        value=value,
                        target=target,
                        func_set=_set_dict,
                        definitions=definitions,
                        queue=queue,
                    )
                continue

            additional_properties = _get_definition(
                definitions=definitions,
                schema=schema.get("additionalProperties", {}),
            )
            if additional_properties:
                for prop_name in source.keys():
                    value = source.get(prop_name)
                    _process_prop(
                        prop_data=additional_properties,
                        prop=prop_name,
                        value=value,
                        target=target,
                        func_set=_set_dict,
                        definitions=definitions,
                        queue=queue,
                    )
                continue

            target.update(source.as_dict())  # type: ignore

        elif type_ == "array":
            if not isinstance(source, simdjson.Array):
                raise ValueError(
                    f"Supposed to be an array, but in reality is a {source.__class__}"
                )

            for i, value in enumerate(source):
                _process_prop(
                    prop_data=_get_definition(definitions, schema["items"]),
                    prop=i,
                    value=value,
                    target=target,
                    func_set=_set_list,
                    definitions=definitions,
                    queue=queue,
                )
        else:
            raise ValueError(f"Invalid schema type {type_}, expected object or array")

    return res


def loads(
    data: Union[str, bytes, bytearray, memoryview],
    *,
    schema: Schema,
    parser: Optional[Parser] = None,
    **_: Any,
) -> JsonType:
    if isinstance(data, str):
        data = data.encode()
    parser = parser or Parser()  # Default for thread safety
    return _loads(data, schema=schema, parser=parser)
